"""This module contains a Fitzhughnagumo cardiac cell model

The module was autogenerated from a gotran form file
"""
from __future__ import division
from collections import OrderedDict
import ufl
import dolfin

from cbcbeat.dolfinimport import backend
from cbcbeat.cellmodels.cardiaccellmodel import CardiacCellModel


class Fitzhughnagumo(CardiacCellModel):
    """
    NOT_IMPLEMENTED
    """

    def __init__(self, params=None, init_conditions=None):
        """
        Create cardiac cell model

        *Arguments*
         params (dict, :py:class:`dolfin.Mesh`, optional)
           optional model parameters
         init_conditions (dict, :py:class:`dolfin.Mesh`, optional)
           optional initial conditions
        """
        CardiacCellModel.__init__(self, params, init_conditions)

    @staticmethod
    def default_parameters():
        "Set-up and return default parameters."
        params = OrderedDict(
            [
                ("a", 0.13),
                ("b", 0.013),
                ("c_1", 0.26),
                ("c_2", 0.1),
                ("c_3", 1.0),
                ("stim_amplitude", 0),
                ("stim_duration", 1),
                ("stim_period", 1000),
                ("stim_start", 1),
                ("v_peak", 40.0),
                ("v_rest", -85.0),
            ]
        )
        return params

    @staticmethod
    def default_initial_conditions():
        "Set-up and return default initial conditions."
        ic = OrderedDict([("v", -85.0), ("s", 0.0)])
        return ic

    def I(self, v, s, time=None):
        """
        Transmembrane current
        """
        # Imports
        # No imports for now

        time = time if time else backend.Constant(0.0)
        # Assign states
        _is_vector = False

        # Assign parameters
        a = self._parameters["a"]
        stim_start = self._parameters["stim_start"]
        stim_amplitude = self._parameters["stim_amplitude"]
        c_1 = self._parameters["c_1"]
        c_2 = self._parameters["c_2"]
        v_rest = self._parameters["v_rest"]
        stim_duration = self._parameters["stim_duration"]
        v_peak = self._parameters["v_peak"]

        current = (
            -(v - v_rest)
            * (v_peak - v)
            * (-(v_peak - v_rest) * a + v - v_rest)
            * c_1
            / ((v_peak - v_rest) * (v_peak - v_rest))
            + (v - v_rest) * c_2 * s / (v_peak - v_rest)
            - (1.0 - 1.0 / (1.0 + ufl.exp(-5.0 * stim_start + 5.0 * time)))
            * stim_amplitude
            / (1.0 + ufl.exp(-5.0 * stim_start + 5.0 * time - 5.0 * stim_duration))
        )

        return current

    def F(self, v, s, time=None):
        """
        Right hand side for ODE system
        """
        # Imports
        # No imports for now

        time = time if time else backend.Constant(0.0)
        # Assign states
        _is_vector = False

        # Assign parameters
        c_3 = self._parameters["c_3"]
        b = self._parameters["b"]
        v_rest = self._parameters["v_rest"]

        F_expressions = [
            # Derivative for state s
            (-c_3 * s + v - v_rest)
            * b,
        ]

        return dolfin.as_vector(F_expressions) if _is_vector else F_expressions[0]

    def num_states(self):
        return 1

    def __str__(self):
        return "Fitzhughnagumo cardiac cell model"



from cbcbeat.cellmodels.beeler_reuter_1977 import Beeler_reuter_1977
from cbcbeat.cellmodels.fitzhughnagumo_manual import FitzHughNagumoManual
from cbcbeat.cellmodels.nocellmodel import NoCellModel
from cbcbeat.cellmodels.rogers_mcculloch_manual import RogersMcCulloch
from cbcbeat.cellmodels.tentusscher_2004_mcell import Tentusscher_2004_mcell
from cbcbeat.cellmodels.tentusscher_panfilov_2006_epi_cell import (
    Tentusscher_panfilov_2006_epi_cell,
)
from cbcbeat.cellmodels.tentusscher_panfilov_2006_M_cell import (
    Tentusscher_panfilov_2006_M_cell,
)
from cbcbeat.cellmodels.fenton_karma_1998_BR_altered import Fenton_karma_1998_BR_altered
from cbcbeat.cellmodels.fenton_karma_1998_MLR1_altered import (
    Fenton_karma_1998_MLR1_altered,
)
from cbcbeat.cellmodels.cardiaccellmodel import CardiacCellModel, MultiCellModel

supported_cell_models = (
    FitzHughNagumoManual,
    NoCellModel,
    RogersMcCulloch,
    Beeler_reuter_1977,
    Tentusscher_2004_mcell,
    Tentusscher_panfilov_2006_epi_cell,
    Fenton_karma_1998_MLR1_altered,
    Fenton_karma_1998_BR_altered,
)


__all__ = [
    "FitzHughNagumoManual",
    "NoCellModel",
    "RogersMcCulloch",
    "Beeler_reuter_1977",
    "Tentusscher_2004_mcell",
    "Tentusscher_panfilov_2006_epi_cell",
    "Fenton_karma_1998_MLR1_altered",
    "Fenton_karma_1998_BR_altered",
    "CardiacCellModel",
    "MultiCellModel",
    "Tentusscher_panfilov_2006_M_cell",
]



"""This module contains a Grandi_pasqualini_bers_2010 cardiac cell model

The module was autogenerated from a gotran ode file
"""
from __future__ import division
from collections import OrderedDict
import ufl
import dolfin

from cbcbeat.dolfinimport import backend
from cbcbeat.cellmodels.cardiaccellmodel import CardiacCellModel


class Grandi_pasqualini_bers_2010(CardiacCellModel):
    def __init__(self, params=None, init_conditions=None):
        """
        Create cardiac cell model

        *Arguments*
         params (dict, :py:class:`dolfin.Mesh`, optional)
           optional model parameters
         init_conditions (dict, :py:class:`dolfin.Mesh`, optional)
           optional initial conditions
        """
        CardiacCellModel.__init__(self, params, init_conditions)

    @staticmethod
    def default_parameters():
        "Set-up and return default parameters."
        params = OrderedDict(
            [
                ("Fjunc", 0.11),
                ("Fjunc_CaL", 0.9),
                ("cellLength", 100),
                ("cellRadius", 10.25),
                ("distJuncSL", 0.5),
                ("distSLcyto", 0.45),
                ("junctionLength", 0.16),
                ("junctionRadius", 0.015),
                ("GNa", 23),
                ("GNaB", 0.000597),
                ("IbarNaK", 1.8),
                ("KmKo", 1.5),
                ("KmNaip", 11),
                ("Q10KmNai", 1.39),
                ("Q10NaK", 1.63),
                ("gkp", 0.002),
                ("pNaK", 0.01833),
                ("epi", 1),
                ("GClB", 0.009),
                ("GClCa", 0.0548125),
                ("KdClCa", 0.1),
                ("Q10CaL", 1.8),
                ("pCa", 0.00027),
                ("pK", 1.35e-07),
                ("pNa", 7.5e-09),
                ("IbarNCX", 4.5),
                ("Kdact", 0.00015),
                ("KmCai", 0.00359),
                ("KmCao", 1.3),
                ("KmNai", 12.29),
                ("KmNao", 87.5),
                ("Q10NCX", 1.57),
                ("ksat", 0.32),
                ("nu", 0.27),
                ("IbarSLCaP", 0.0673),
                ("KmPCa", 0.0005),
                ("Q10SLCaP", 2.35),
                ("GCaB", 0.0005513),
                ("Kmf", 0.000246),
                ("Kmr", 1.7),
                ("MaxSR", 15),
                ("MinSR", 1),
                ("Q10SRCaP", 2.6),
                ("Vmax_SRCaP", 0.0053114),
                ("ec50SR", 0.45),
                ("hillSRCaP", 1.787),
                ("kiCa", 0.5),
                ("kim", 0.005),
                ("koCa", 10),
                ("kom", 0.06),
                ("ks", 25),
                ("Bmax_Naj", 7.561),
                ("Bmax_Nasl", 1.65),
                ("koff_na", 0.001),
                ("kon_na", 0.0001),
                ("Bmax_CaM", 0.024),
                ("Bmax_SR", 0.0171),
                ("Bmax_TnChigh", 0.14),
                ("Bmax_TnClow", 0.07),
                ("Bmax_myosin", 0.14),
                ("koff_cam", 0.238),
                ("koff_myoca", 0.00046),
                ("koff_myomg", 5.7e-05),
                ("koff_sr", 0.06),
                ("koff_tnchca", 3.2e-05),
                ("koff_tnchmg", 0.00333),
                ("koff_tncl", 0.0196),
                ("kon_cam", 34),
                ("kon_myoca", 13.8),
                ("kon_myomg", 0.0157),
                ("kon_sr", 100),
                ("kon_tnchca", 2.37),
                ("kon_tnchmg", 0.003),
                ("kon_tncl", 32.7),
                ("Bmax_SLhighj0", 0.000165),
                ("Bmax_SLhighsl0", 0.0134),
                ("Bmax_SLlowj0", 0.00046),
                ("Bmax_SLlowsl0", 0.0374),
                ("koff_slh", 0.03),
                ("koff_sll", 1.3),
                ("kon_slh", 100),
                ("kon_sll", 100),
                ("Bmax_Csqn0", 0.14),
                ("DcaJuncSL", 1.64e-06),
                ("DcaSLcyto", 1.22e-06),
                ("J_ca_juncsl", 8.2413e-13),
                ("J_ca_slmyo", 3.7243e-12),
                ("koff_csqn", 65),
                ("kon_csqn", 100),
                ("DnaJuncSL", 1.09e-05),
                ("DnaSLcyto", 1.79e-05),
                ("J_na_juncsl", 1.8313e-14),
                ("J_na_slmyo", 1.6386e-12),
                ("Nao", 140),
                ("Ko", 5.4),
                ("Cao", 1.8),
                ("Cli", 15),
                ("Clo", 150),
                ("Mgi", 1),
                ("Cmem", 1.381e-10),
                ("Frdy", 96485),
                ("R", 8314),
                ("Temp", 310),
                ("stim_amplitude", 0.0),
                ("stim_duration", 1.0),
                ("stim_period", 1000.0),
                ("stim_start", 1.0),
            ]
        )
        return params

    @staticmethod
    def default_initial_conditions():
        "Set-up and return default initial conditions."
        ic = OrderedDict(
            [
                ("V_m", -81.4552030513),
                ("h", 0.626221949492),
                ("j", 0.62455357249),
                ("m", 0.00379308741444),
                ("x_kr", 0.0210022533039),
                ("x_ks", 0.00428016666259),
                ("x_to_f", 0.000440438103759),
                ("x_to_s", 0.000440445885643),
                ("y_to_f", 0.999995844039),
                ("y_to_s", 0.785115828275),
                ("d", 2.92407183949e-06),
                ("f", 0.995135796704),
                ("f_Ca_Bj", 0.0246760872106),
                ("f_Ca_Bsl", 0.0152723084239),
                ("Ry_Ri", 9.07666168961e-08),
                ("Ry_Ro", 7.40481128854e-07),
                ("Ry_Rr", 0.890806040818),
                ("Na_Bj", 3.45437733033),
                ("Na_Bsl", 0.753740951478),
                ("CaM", 0.000295573424135),
                ("Myo_c", 0.00192322252438),
                ("Myo_m", 0.137560495023),
                ("SRB", 0.00217360235649),
                ("Tn_CHc", 0.117412025937),
                ("Tn_CHm", 0.0106160166693),
                ("Tn_CL", 0.00893455096919),
                ("SLH_j", 0.0735890020284),
                ("SLH_sl", 0.114583623437),
                ("SLL_j", 0.0074052452168),
                ("SLL_sl", 0.00990339304377),
                ("Ca_sr", 0.554760499828),
                ("Csqn_b", 1.19723145924),
                ("Na_i", 8.40513364345),
                ("Na_j", 8.40537012593),
                ("Na_sl", 8.40491910001),
                ("K_i", 120),
                ("Ca_i", 8.72509677797e-05),
                ("Ca_j", 0.000175882395147),
                ("Ca_sl", 0.000106779509977),
            ]
        )
        return ic

    def _I(self, v, s, time):
        """
        Original gotran transmembrane current dV/dt
        """
        time = time if time else backend.Constant(0.0)

        # Assign states
        V_m = v
        assert len(s) == 38
        (
            h,
            j,
            m,
            x_kr,
            x_ks,
            x_to_f,
            x_to_s,
            y_to_f,
            y_to_s,
            d,
            f,
            f_Ca_Bj,
            f_Ca_Bsl,
            Ry_Ri,
            Ry_Ro,
            Ry_Rr,
            Na_Bj,
            Na_Bsl,
            CaM,
            Myo_c,
            Myo_m,
            SRB,
            Tn_CHc,
            Tn_CHm,
            Tn_CL,
            SLH_j,
            SLH_sl,
            SLL_j,
            SLL_sl,
            Ca_sr,
            Csqn_b,
            Na_i,
            Na_j,
            Na_sl,
            K_i,
            Ca_i,
            Ca_j,
            Ca_sl,
        ) = s

        # Assign parameters
        Fjunc = self._parameters["Fjunc"]
        Fjunc_CaL = self._parameters["Fjunc_CaL"]
        GNa = self._parameters["GNa"]
        GNaB = self._parameters["GNaB"]
        IbarNaK = self._parameters["IbarNaK"]
        KmKo = self._parameters["KmKo"]
        KmNaip = self._parameters["KmNaip"]
        gkp = self._parameters["gkp"]
        pNaK = self._parameters["pNaK"]
        epi = self._parameters["epi"]
        GClB = self._parameters["GClB"]
        GClCa = self._parameters["GClCa"]
        KdClCa = self._parameters["KdClCa"]
        Q10CaL = self._parameters["Q10CaL"]
        pCa = self._parameters["pCa"]
        pK = self._parameters["pK"]
        pNa = self._parameters["pNa"]
        IbarNCX = self._parameters["IbarNCX"]
        Kdact = self._parameters["Kdact"]
        KmCai = self._parameters["KmCai"]
        KmCao = self._parameters["KmCao"]
        KmNai = self._parameters["KmNai"]
        KmNao = self._parameters["KmNao"]
        Q10NCX = self._parameters["Q10NCX"]
        ksat = self._parameters["ksat"]
        nu = self._parameters["nu"]
        IbarSLCaP = self._parameters["IbarSLCaP"]
        KmPCa = self._parameters["KmPCa"]
        Q10SLCaP = self._parameters["Q10SLCaP"]
        GCaB = self._parameters["GCaB"]
        Nao = self._parameters["Nao"]
        Ko = self._parameters["Ko"]
        Cao = self._parameters["Cao"]
        Cli = self._parameters["Cli"]
        Clo = self._parameters["Clo"]
        Frdy = self._parameters["Frdy"]
        R = self._parameters["R"]
        Temp = self._parameters["Temp"]

        # Init return args
        current = [ufl.zero()] * 1

        # Expressions for the Geometry component
        Fsl = 1 - Fjunc
        Fsl_CaL = 1 - Fjunc_CaL

        # Expressions for the Reversal potentials component
        FoRT = Frdy / (R * Temp)
        ena_junc = ufl.ln(Nao / Na_j) / FoRT
        ena_sl = ufl.ln(Nao / Na_sl) / FoRT
        ek = ufl.ln(Ko / K_i) / FoRT
        eca_junc = ufl.ln(Cao / Ca_j) / (2 * FoRT)
        eca_sl = ufl.ln(Cao / Ca_sl) / (2 * FoRT)
        ecl = ufl.ln(Cli / Clo) / FoRT
        Qpow = -31 + Temp / 10

        # Expressions for the I_Na component
        I_Na_junc = Fjunc * GNa * (m * m * m) * (-ena_junc + V_m) * h * j
        I_Na_sl = GNa * (m * m * m) * (-ena_sl + V_m) * Fsl * h * j

        # Expressions for the I_NaBK component
        I_nabk_junc = Fjunc * GNaB * (-ena_junc + V_m)
        I_nabk_sl = GNaB * (-ena_sl + V_m) * Fsl

        # Expressions for the I_NaK component
        sigma = -1 / 7 + ufl.exp(0.0148588410104 * Nao) / 7
        fnak = 1.0 / (
            1
            + 0.1245 * ufl.exp(-0.1 * FoRT * V_m)
            + 0.0365 * ufl.exp(-FoRT * V_m) * sigma
        )
        I_nak_junc = (
            Fjunc
            * IbarNaK
            * Ko
            * fnak
            / ((1 + ufl.elem_pow(KmNaip, 4) / ufl.elem_pow(Na_j, 4)) * (KmKo + Ko))
        )
        I_nak_sl = (
            IbarNaK
            * Ko
            * Fsl
            * fnak
            / ((1 + ufl.elem_pow(KmNaip, 4) / ufl.elem_pow(Na_sl, 4)) * (KmKo + Ko))
        )
        I_nak = I_nak_junc + I_nak_sl

        # Expressions for the I_Kr component
        gkr = 0.0150616019019 * ufl.sqrt(Ko)
        rkr = 1.0 / (1 + ufl.exp(37 / 12 + V_m / 24))
        I_kr = (-ek + V_m) * gkr * rkr * x_kr

        # Expressions for the I_Kp component
        kp_kp = 1.0 / (1 + 1786.47556538 * ufl.exp(-0.167224080268 * V_m))
        I_kp_junc = Fjunc * gkp * (-ek + V_m) * kp_kp
        I_kp_sl = gkp * (-ek + V_m) * Fsl * kp_kp
        I_kp = I_kp_sl + I_kp_junc

        # Expressions for the I_Ks component
        eks = ufl.ln((Nao * pNaK + Ko) / (pNaK * Na_i + K_i)) / FoRT
        gks_junc = 0.0035
        gks_sl = 0.0035
        I_ks_junc = Fjunc * gks_junc * (x_ks * x_ks) * (-eks + V_m)
        I_ks_sl = gks_sl * (x_ks * x_ks) * (-eks + V_m) * Fsl
        I_ks = I_ks_sl + I_ks_junc

        # Expressions for the I_to component
        GtoSlow = ufl.conditional(ufl.eq(epi, 1), 0.0156, 0.037596)
        GtoFast = ufl.conditional(ufl.eq(epi, 1), 0.1144, 0.001404)
        I_tos = (-ek + V_m) * GtoSlow * x_to_s * y_to_s
        I_tof = (-ek + V_m) * GtoFast * x_to_f * y_to_f
        I_to = I_tos + I_tof

        # Expressions for the I_Ki component
        aki = 1.02 / (1 + 7.35454251046e-07 * ufl.exp(0.2385 * V_m - 0.2385 * ek))
        bki = (
            0.762624006506 * ufl.exp(0.08032 * V_m - 0.08032 * ek)
            + 1.15340563519e-16 * ufl.exp(0.06175 * V_m - 0.06175 * ek)
        ) / (1 + 0.0867722941577 * ufl.exp(-0.5143 * V_m + 0.5143 * ek))
        kiss = aki / (aki + bki)
        I_ki = 0.150616019019 * ufl.sqrt(Ko) * (-ek + V_m) * kiss

        # Expressions for the I_ClCa component
        I_ClCa_junc = Fjunc * GClCa * (-ecl + V_m) / (1 + KdClCa / Ca_j)
        I_ClCa_sl = GClCa * (-ecl + V_m) * Fsl / (1 + KdClCa / Ca_sl)
        I_ClCa = I_ClCa_sl + I_ClCa_junc
        I_Clbk = GClB * (-ecl + V_m)

        # Expressions for the I_Ca component
        fcaCaMSL = 0
        fcaCaj = 0
        ibarca_j = (
            4
            * Frdy
            * pCa
            * (-0.341 * Cao + 0.341 * Ca_j * ufl.exp(2 * FoRT * V_m))
            * FoRT
            * V_m
            / (-1 + ufl.exp(2 * FoRT * V_m))
        )
        ibarca_sl = (
            4
            * Frdy
            * pCa
            * (-0.341 * Cao + 0.341 * Ca_sl * ufl.exp(2 * FoRT * V_m))
            * FoRT
            * V_m
            / (-1 + ufl.exp(2 * FoRT * V_m))
        )
        ibark = (
            Frdy
            * pK
            * (-0.75 * Ko + 0.75 * K_i * ufl.exp(FoRT * V_m))
            * FoRT
            * V_m
            / (-1 + ufl.exp(FoRT * V_m))
        )
        ibarna_j = (
            Frdy
            * pNa
            * (-0.75 * Nao + 0.75 * Na_j * ufl.exp(FoRT * V_m))
            * FoRT
            * V_m
            / (-1 + ufl.exp(FoRT * V_m))
        )
        ibarna_sl = (
            Frdy
            * pNa
            * (0.75 * Na_sl * ufl.exp(FoRT * V_m) - 0.75 * Nao)
            * FoRT
            * V_m
            / (-1 + ufl.exp(FoRT * V_m))
        )
        I_Ca_junc = (
            0.45
            * Fjunc_CaL
            * ufl.elem_pow(Q10CaL, Qpow)
            * (1 - f_Ca_Bj + fcaCaj)
            * d
            * f
            * ibarca_j
        )
        I_Ca_sl = (
            0.45
            * ufl.elem_pow(Q10CaL, Qpow)
            * (1 - f_Ca_Bsl + fcaCaMSL)
            * Fsl_CaL
            * d
            * f
            * ibarca_sl
        )
        I_CaK = (
            0.45
            * ufl.elem_pow(Q10CaL, Qpow)
            * (Fjunc_CaL * (1 - f_Ca_Bj + fcaCaj) + (1 - f_Ca_Bsl + fcaCaMSL) * Fsl_CaL)
            * d
            * f
            * ibark
        )
        I_CaNa_junc = (
            0.45
            * Fjunc_CaL
            * ufl.elem_pow(Q10CaL, Qpow)
            * (1 - f_Ca_Bj + fcaCaj)
            * d
            * f
            * ibarna_j
        )
        I_CaNa_sl = (
            0.45
            * ufl.elem_pow(Q10CaL, Qpow)
            * (1 - f_Ca_Bsl + fcaCaMSL)
            * Fsl_CaL
            * d
            * f
            * ibarna_sl
        )

        # Expressions for the I_NCX component
        Ka_junc = 1.0 / (1 + (Kdact * Kdact) / (Ca_j * Ca_j))
        Ka_sl = 1.0 / (1 + (Kdact * Kdact) / (Ca_sl * Ca_sl))
        s1_junc = Cao * (Na_j * Na_j * Na_j) * ufl.exp(nu * FoRT * V_m)
        s1_sl = Cao * (Na_sl * Na_sl * Na_sl) * ufl.exp(nu * FoRT * V_m)
        s2_junc = (Nao * Nao * Nao) * Ca_j * ufl.exp((-1 + nu) * FoRT * V_m)
        s3_junc = (
            KmCao * (Na_j * Na_j * Na_j)
            + (Nao * Nao * Nao) * Ca_j
            + Cao * (Na_j * Na_j * Na_j)
            + KmCai
            * (Nao * Nao * Nao)
            * (1 + (Na_j * Na_j * Na_j) / (KmNai * KmNai * KmNai))
            + (KmNao * KmNao * KmNao) * (1 + Ca_j / KmCai) * Ca_j
        )
        s2_sl = (Nao * Nao * Nao) * Ca_sl * ufl.exp((-1 + nu) * FoRT * V_m)
        s3_sl = (
            KmCai
            * (Nao * Nao * Nao)
            * (1 + (Na_sl * Na_sl * Na_sl) / (KmNai * KmNai * KmNai))
            + (Nao * Nao * Nao) * Ca_sl
            + (KmNao * KmNao * KmNao) * (1 + Ca_sl / KmCai) * Ca_sl
            + Cao * (Na_sl * Na_sl * Na_sl)
            + KmCao * (Na_sl * Na_sl * Na_sl)
        )
        I_ncx_junc = (
            Fjunc
            * IbarNCX
            * ufl.elem_pow(Q10NCX, Qpow)
            * (-s2_junc + s1_junc)
            * Ka_junc
            / ((1 + ksat * ufl.exp((-1 + nu) * FoRT * V_m)) * s3_junc)
        )
        I_ncx_sl = (
            IbarNCX
            * ufl.elem_pow(Q10NCX, Qpow)
            * (-s2_sl + s1_sl)
            * Fsl
            * Ka_sl
            / ((1 + ksat * ufl.exp((-1 + nu) * FoRT * V_m)) * s3_sl)
        )

        # Expressions for the I_PCa component
        I_pca_junc = (
            Fjunc
            * IbarSLCaP
            * ufl.elem_pow(Q10SLCaP, Qpow)
            * ufl.elem_pow(Ca_j, 1.6)
            / (ufl.elem_pow(Ca_j, 1.6) + ufl.elem_pow(KmPCa, 1.6))
        )
        I_pca_sl = (
            IbarSLCaP
            * ufl.elem_pow(Q10SLCaP, Qpow)
            * ufl.elem_pow(Ca_sl, 1.6)
            * Fsl
            / (ufl.elem_pow(Ca_sl, 1.6) + ufl.elem_pow(KmPCa, 1.6))
        )

        # Expressions for the I_CaBK component
        I_cabk_junc = Fjunc * GCaB * (-eca_junc + V_m)
        I_cabk_sl = GCaB * (-eca_sl + V_m) * Fsl

        # Expressions for the Na Concentrations component
        I_Na_tot_junc = (
            3 * I_nak_junc + 3 * I_ncx_junc + I_CaNa_junc + I_Na_junc + I_nabk_junc
        )
        I_Na_tot_sl = I_Na_sl + I_nabk_sl + 3 * I_nak_sl + I_CaNa_sl + 3 * I_ncx_sl

        # Expressions for the K Concentration component
        I_K_tot = -2 * I_nak + I_ks + I_CaK + I_kr + I_kp + I_ki + I_to

        # Expressions for the Ca Concentrations component
        I_Ca_tot_junc = I_pca_junc + I_cabk_junc + I_Ca_junc - 2 * I_ncx_junc
        I_Ca_tot_sl = -2 * I_ncx_sl + I_pca_sl + I_cabk_sl + I_Ca_sl

        # Expressions for the Membrane potential component
        i_Stim = 0
        I_Na_tot = I_Na_tot_junc + I_Na_tot_sl
        I_Cl_tot = I_Clbk + I_ClCa
        I_Ca_tot = I_Ca_tot_junc + I_Ca_tot_sl
        I_tot = I_Na_tot + I_K_tot + I_Cl_tot + I_Ca_tot
        current[0] = -I_tot - i_Stim

        # Return results
        return current[0]

    def I(self, v, s, time=None):
        """
        Transmembrane current

           I = -dV/dt

        """
        return -self._I(v, s, time)

    def F(self, v, s, time=None):
        """
        Right hand side for ODE system
        """
        time = time if time else backend.Constant(0.0)

        # Assign states
        V_m = v
        assert len(s) == 38
        (
            h,
            j,
            m,
            x_kr,
            x_ks,
            x_to_f,
            x_to_s,
            y_to_f,
            y_to_s,
            d,
            f,
            f_Ca_Bj,
            f_Ca_Bsl,
            Ry_Ri,
            Ry_Ro,
            Ry_Rr,
            Na_Bj,
            Na_Bsl,
            CaM,
            Myo_c,
            Myo_m,
            SRB,
            Tn_CHc,
            Tn_CHm,
            Tn_CL,
            SLH_j,
            SLH_sl,
            SLL_j,
            SLL_sl,
            Ca_sr,
            Csqn_b,
            Na_i,
            Na_j,
            Na_sl,
            K_i,
            Ca_i,
            Ca_j,
            Ca_sl,
        ) = s

        # Assign parameters
        Fjunc = self._parameters["Fjunc"]
        Fjunc_CaL = self._parameters["Fjunc_CaL"]
        cellLength = self._parameters["cellLength"]
        cellRadius = self._parameters["cellRadius"]
        GNa = self._parameters["GNa"]
        GNaB = self._parameters["GNaB"]
        IbarNaK = self._parameters["IbarNaK"]
        KmKo = self._parameters["KmKo"]
        KmNaip = self._parameters["KmNaip"]
        Q10CaL = self._parameters["Q10CaL"]
        pCa = self._parameters["pCa"]
        pNa = self._parameters["pNa"]
        IbarNCX = self._parameters["IbarNCX"]
        Kdact = self._parameters["Kdact"]
        KmCai = self._parameters["KmCai"]
        KmCao = self._parameters["KmCao"]
        KmNai = self._parameters["KmNai"]
        KmNao = self._parameters["KmNao"]
        Q10NCX = self._parameters["Q10NCX"]
        ksat = self._parameters["ksat"]
        nu = self._parameters["nu"]
        IbarSLCaP = self._parameters["IbarSLCaP"]
        KmPCa = self._parameters["KmPCa"]
        Q10SLCaP = self._parameters["Q10SLCaP"]
        GCaB = self._parameters["GCaB"]
        Kmf = self._parameters["Kmf"]
        Kmr = self._parameters["Kmr"]
        MaxSR = self._parameters["MaxSR"]
        MinSR = self._parameters["MinSR"]
        Q10SRCaP = self._parameters["Q10SRCaP"]
        Vmax_SRCaP = self._parameters["Vmax_SRCaP"]
        ec50SR = self._parameters["ec50SR"]
        hillSRCaP = self._parameters["hillSRCaP"]
        kiCa = self._parameters["kiCa"]
        kim = self._parameters["kim"]
        koCa = self._parameters["koCa"]
        kom = self._parameters["kom"]
        ks = self._parameters["ks"]
        Bmax_Naj = self._parameters["Bmax_Naj"]
        Bmax_Nasl = self._parameters["Bmax_Nasl"]
        koff_na = self._parameters["koff_na"]
        kon_na = self._parameters["kon_na"]
        Bmax_CaM = self._parameters["Bmax_CaM"]
        Bmax_SR = self._parameters["Bmax_SR"]
        Bmax_TnChigh = self._parameters["Bmax_TnChigh"]
        Bmax_TnClow = self._parameters["Bmax_TnClow"]
        Bmax_myosin = self._parameters["Bmax_myosin"]
        koff_cam = self._parameters["koff_cam"]
        koff_myoca = self._parameters["koff_myoca"]
        koff_myomg = self._parameters["koff_myomg"]
        koff_sr = self._parameters["koff_sr"]
        koff_tnchca = self._parameters["koff_tnchca"]
        koff_tnchmg = self._parameters["koff_tnchmg"]
        koff_tncl = self._parameters["koff_tncl"]
        kon_cam = self._parameters["kon_cam"]
        kon_myoca = self._parameters["kon_myoca"]
        kon_myomg = self._parameters["kon_myomg"]
        kon_sr = self._parameters["kon_sr"]
        kon_tnchca = self._parameters["kon_tnchca"]
        kon_tnchmg = self._parameters["kon_tnchmg"]
        kon_tncl = self._parameters["kon_tncl"]
        Bmax_SLhighj0 = self._parameters["Bmax_SLhighj0"]
        Bmax_SLhighsl0 = self._parameters["Bmax_SLhighsl0"]
        Bmax_SLlowj0 = self._parameters["Bmax_SLlowj0"]
        Bmax_SLlowsl0 = self._parameters["Bmax_SLlowsl0"]
        koff_slh = self._parameters["koff_slh"]
        koff_sll = self._parameters["koff_sll"]
        kon_slh = self._parameters["kon_slh"]
        kon_sll = self._parameters["kon_sll"]
        Bmax_Csqn0 = self._parameters["Bmax_Csqn0"]
        J_ca_juncsl = self._parameters["J_ca_juncsl"]
        J_ca_slmyo = self._parameters["J_ca_slmyo"]
        koff_csqn = self._parameters["koff_csqn"]
        kon_csqn = self._parameters["kon_csqn"]
        J_na_juncsl = self._parameters["J_na_juncsl"]
        J_na_slmyo = self._parameters["J_na_slmyo"]
        Nao = self._parameters["Nao"]
        Ko = self._parameters["Ko"]
        Cao = self._parameters["Cao"]
        Mgi = self._parameters["Mgi"]
        Cmem = self._parameters["Cmem"]
        Frdy = self._parameters["Frdy"]
        R = self._parameters["R"]
        Temp = self._parameters["Temp"]

        # Init return args
        F_expressions = [ufl.zero()] * 38

        # Expressions for the Geometry component
        Vcell = 1e-15 * ufl.pi * cellLength * (cellRadius * cellRadius)
        Vmyo = 0.65 * Vcell
        Vsr = 0.035 * Vcell
        Vsl = 0.02 * Vcell
        Vjunc = 0.000539 * Vcell
        Fsl = 1 - Fjunc
        Fsl_CaL = 1 - Fjunc_CaL

        # Expressions for the Reversal potentials component
        FoRT = Frdy / (R * Temp)
        ena_junc = ufl.ln(Nao / Na_j) / FoRT
        ena_sl = ufl.ln(Nao / Na_sl) / FoRT
        eca_junc = ufl.ln(Cao / Ca_j) / (2 * FoRT)
        eca_sl = ufl.ln(Cao / Ca_sl) / (2 * FoRT)
        Qpow = -31 + Temp / 10

        # Expressions for the I_Na component
        mss = 1.0 / (
            (1 + 0.00184221158117 * ufl.exp(-0.110741971207 * V_m))
            * (1 + 0.00184221158117 * ufl.exp(-0.110741971207 * V_m))
        )
        taum = 0.1292 * ufl.exp(
            -(
                (2.94658944659 + 0.0643500643501 * V_m)
                * (2.94658944659 + 0.0643500643501 * V_m)
            )
        ) + 0.06487 * ufl.exp(
            -(
                (-0.0943466353678 + 0.0195618153365 * V_m)
                * (-0.0943466353678 + 0.0195618153365 * V_m)
            )
        )
        ah = ufl.conditional(
            ufl.ge(V_m, -40), 0, 4.43126792958e-07 * ufl.exp(-0.147058823529 * V_m)
        )
        bh = ufl.conditional(
            ufl.ge(V_m, -40),
            0.77 / (0.13 + 0.0497581410839 * ufl.exp(-0.0900900900901 * V_m)),
            310000.0 * ufl.exp(0.3485 * V_m) + 2.7 * ufl.exp(0.079 * V_m),
        )
        tauh = 1.0 / (bh + ah)
        hss = 1.0 / (
            (1 + 15212.5932857 * ufl.exp(0.134589502019 * V_m))
            * (1 + 15212.5932857 * ufl.exp(0.134589502019 * V_m))
        )
        aj = ufl.conditional(
            ufl.ge(V_m, -40),
            0,
            (37.78 + V_m)
            * (-25428.0 * ufl.exp(0.2444 * V_m) - 6.948e-06 * ufl.exp(-0.04391 * V_m))
            / (1 + 50262745826.0 * ufl.exp(0.311 * V_m)),
        )
        bj = ufl.conditional(
            ufl.ge(V_m, -40),
            0.6 * ufl.exp(0.057 * V_m) / (1 + 0.0407622039784 * ufl.exp(-0.1 * V_m)),
            0.02424
            * ufl.exp(-0.01052 * V_m)
            / (1 + 0.0039608683399 * ufl.exp(-0.1378 * V_m)),
        )
        tauj = 1.0 / (bj + aj)
        jss = 1.0 / (
            (1 + 15212.5932857 * ufl.exp(0.134589502019 * V_m))
            * (1 + 15212.5932857 * ufl.exp(0.134589502019 * V_m))
        )
        F_expressions[2] = (-m + mss) / taum
        F_expressions[0] = (hss - h) / tauh
        F_expressions[1] = (-j + jss) / tauj
        I_Na_junc = Fjunc * GNa * (m * m * m) * (-ena_junc + V_m) * h * j
        I_Na_sl = GNa * (m * m * m) * (-ena_sl + V_m) * Fsl * h * j

        # Expressions for the I_NaBK component
        I_nabk_junc = Fjunc * GNaB * (-ena_junc + V_m)
        I_nabk_sl = GNaB * (-ena_sl + V_m) * Fsl

        # Expressions for the I_NaK component
        sigma = -1 / 7 + ufl.exp(0.0148588410104 * Nao) / 7
        fnak = 1.0 / (
            1
            + 0.1245 * ufl.exp(-0.1 * FoRT * V_m)
            + 0.0365 * ufl.exp(-FoRT * V_m) * sigma
        )
        I_nak_junc = (
            Fjunc
            * IbarNaK
            * Ko
            * fnak
            / ((1 + ufl.elem_pow(KmNaip, 4) / ufl.elem_pow(Na_j, 4)) * (KmKo + Ko))
        )
        I_nak_sl = (
            IbarNaK
            * Ko
            * Fsl
            * fnak
            / ((1 + ufl.elem_pow(KmNaip, 4) / ufl.elem_pow(Na_sl, 4)) * (KmKo + Ko))
        )

        # Expressions for the I_Kr component
        xrss = 1.0 / (1 + ufl.exp(-2 - V_m / 5))
        tauxr = 230 / (1 + ufl.exp(2 + V_m / 20)) + 3300 / (
            (1 + ufl.exp(-22 / 9 - V_m / 9)) * (1 + ufl.exp(11 / 9 + V_m / 9))
        )
        F_expressions[3] = (-x_kr + xrss) / tauxr

        # Expressions for the I_Ks component
        xsss = 1.0 / (1 + 0.765928338365 * ufl.exp(-0.0701754385965 * V_m))
        tauxs = 990.1 / (1 + 0.841540408868 * ufl.exp(-0.070821529745 * V_m))
        F_expressions[4] = (-x_ks + xsss) / tauxs

        # Expressions for the I_to component
        xtoss = 1.0 / (1 + ufl.exp(19 / 13 - V_m / 13))
        ytoss = 1.0 / (1 + 49.4024491055 * ufl.exp(V_m / 5))
        tauxtos = 0.5 + 9 / (1 + ufl.exp(1 / 5 + V_m / 15))
        tauytos = 30 + 800 / (1 + ufl.exp(6 + V_m / 10))
        F_expressions[6] = (-x_to_s + xtoss) / tauxtos
        F_expressions[8] = (ytoss - y_to_s) / tauytos
        tauxtof = 0.5 + 8.5 * ufl.exp(-((9 / 10 + V_m / 50) * (9 / 10 + V_m / 50)))
        tauytof = 7 + 85 * ufl.exp(-((40 + V_m) * (40 + V_m)) / 220)
        F_expressions[5] = (xtoss - x_to_f) / tauxtof
        F_expressions[7] = (ytoss - y_to_f) / tauytof

        # Expressions for the I_Ca component
        fss = 0.6 / (1 + ufl.exp(5 / 2 - V_m / 20)) + 1.0 / (
            1 + ufl.exp(35 / 9 + V_m / 9)
        )
        dss = 1.0 / (1 + ufl.exp(-5 / 6 - V_m / 6))
        taud = (1 - ufl.exp(-5 / 6 - V_m / 6)) * dss / (0.175 + 0.035 * V_m)
        tauf = 1.0 / (
            0.02
            + 0.0197 * ufl.exp(-((0.48865 + 0.0337 * V_m) * (0.48865 + 0.0337 * V_m)))
        )
        F_expressions[9] = (-d + dss) / taud
        F_expressions[10] = (fss - f) / tauf
        F_expressions[11] = 1.7 * (1 - f_Ca_Bj) * Ca_j - 0.0119 * f_Ca_Bj
        F_expressions[12] = -0.0119 * f_Ca_Bsl + 1.7 * (1 - f_Ca_Bsl) * Ca_sl
        fcaCaMSL = 0
        fcaCaj = 0
        ibarca_j = (
            4
            * Frdy
            * pCa
            * (-0.341 * Cao + 0.341 * Ca_j * ufl.exp(2 * FoRT * V_m))
            * FoRT
            * V_m
            / (-1 + ufl.exp(2 * FoRT * V_m))
        )
        ibarca_sl = (
            4
            * Frdy
            * pCa
            * (-0.341 * Cao + 0.341 * Ca_sl * ufl.exp(2 * FoRT * V_m))
            * FoRT
            * V_m
            / (-1 + ufl.exp(2 * FoRT * V_m))
        )
        ibarna_j = (
            Frdy
            * pNa
            * (-0.75 * Nao + 0.75 * Na_j * ufl.exp(FoRT * V_m))
            * FoRT
            * V_m
            / (-1 + ufl.exp(FoRT * V_m))
        )
        ibarna_sl = (
            Frdy
            * pNa
            * (0.75 * Na_sl * ufl.exp(FoRT * V_m) - 0.75 * Nao)
            * FoRT
            * V_m
            / (-1 + ufl.exp(FoRT * V_m))
        )
        I_Ca_junc = (
            0.45
            * Fjunc_CaL
            * ufl.elem_pow(Q10CaL, Qpow)
            * (1 - f_Ca_Bj + fcaCaj)
            * d
            * f
            * ibarca_j
        )
        I_Ca_sl = (
            0.45
            * ufl.elem_pow(Q10CaL, Qpow)
            * (1 - f_Ca_Bsl + fcaCaMSL)
            * Fsl_CaL
            * d
            * f
            * ibarca_sl
        )
        I_CaNa_junc = (
            0.45
            * Fjunc_CaL
            * ufl.elem_pow(Q10CaL, Qpow)
            * (1 - f_Ca_Bj + fcaCaj)
            * d
            * f
            * ibarna_j
        )
        I_CaNa_sl = (
            0.45
            * ufl.elem_pow(Q10CaL, Qpow)
            * (1 - f_Ca_Bsl + fcaCaMSL)
            * Fsl_CaL
            * d
            * f
            * ibarna_sl
        )

        # Expressions for the I_NCX component
        Ka_junc = 1.0 / (1 + (Kdact * Kdact) / (Ca_j * Ca_j))
        Ka_sl = 1.0 / (1 + (Kdact * Kdact) / (Ca_sl * Ca_sl))
        s1_junc = Cao * (Na_j * Na_j * Na_j) * ufl.exp(nu * FoRT * V_m)
        s1_sl = Cao * (Na_sl * Na_sl * Na_sl) * ufl.exp(nu * FoRT * V_m)
        s2_junc = (Nao * Nao * Nao) * Ca_j * ufl.exp((-1 + nu) * FoRT * V_m)
        s3_junc = (
            KmCao * (Na_j * Na_j * Na_j)
            + (Nao * Nao * Nao) * Ca_j
            + Cao * (Na_j * Na_j * Na_j)
            + KmCai
            * (Nao * Nao * Nao)
            * (1 + (Na_j * Na_j * Na_j) / (KmNai * KmNai * KmNai))
            + (KmNao * KmNao * KmNao) * (1 + Ca_j / KmCai) * Ca_j
        )
        s2_sl = (Nao * Nao * Nao) * Ca_sl * ufl.exp((-1 + nu) * FoRT * V_m)
        s3_sl = (
            KmCai
            * (Nao * Nao * Nao)
            * (1 + (Na_sl * Na_sl * Na_sl) / (KmNai * KmNai * KmNai))
            + (Nao * Nao * Nao) * Ca_sl
            + (KmNao * KmNao * KmNao) * (1 + Ca_sl / KmCai) * Ca_sl
            + Cao * (Na_sl * Na_sl * Na_sl)
            + KmCao * (Na_sl * Na_sl * Na_sl)
        )
        I_ncx_junc = (
            Fjunc
            * IbarNCX
            * ufl.elem_pow(Q10NCX, Qpow)
            * (-s2_junc + s1_junc)
            * Ka_junc
            / ((1 + ksat * ufl.exp((-1 + nu) * FoRT * V_m)) * s3_junc)
        )
        I_ncx_sl = (
            IbarNCX
            * ufl.elem_pow(Q10NCX, Qpow)
            * (-s2_sl + s1_sl)
            * Fsl
            * Ka_sl
            / ((1 + ksat * ufl.exp((-1 + nu) * FoRT * V_m)) * s3_sl)
        )

        # Expressions for the I_PCa component
        I_pca_junc = (
            Fjunc
            * IbarSLCaP
            * ufl.elem_pow(Q10SLCaP, Qpow)
            * ufl.elem_pow(Ca_j, 1.6)
            / (ufl.elem_pow(Ca_j, 1.6) + ufl.elem_pow(KmPCa, 1.6))
        )
        I_pca_sl = (
            IbarSLCaP
            * ufl.elem_pow(Q10SLCaP, Qpow)
            * ufl.elem_pow(Ca_sl, 1.6)
            * Fsl
            / (ufl.elem_pow(Ca_sl, 1.6) + ufl.elem_pow(KmPCa, 1.6))
        )

        # Expressions for the I_CaBK component
        I_cabk_junc = Fjunc * GCaB * (-eca_junc + V_m)
        I_cabk_sl = GCaB * (-eca_sl + V_m) * Fsl

        # Expressions for the SR Fluxes component
        kCaSR = MaxSR - (-MinSR + MaxSR) / (1 + ufl.elem_pow(ec50SR / Ca_sr, 2.5))
        koSRCa = koCa / kCaSR
        kiSRCa = kiCa * kCaSR
        RI = 1 - Ry_Ro - Ry_Ri - Ry_Rr
        F_expressions[15] = (
            -(Ca_j * Ca_j) * Ry_Rr * koSRCa
            + kom * Ry_Ro
            + kim * RI
            - Ca_j * Ry_Rr * kiSRCa
        )
        F_expressions[14] = (
            -kom * Ry_Ro
            - Ca_j * Ry_Ro * kiSRCa
            + kim * Ry_Ri
            + (Ca_j * Ca_j) * Ry_Rr * koSRCa
        )
        F_expressions[13] = (
            -kim * Ry_Ri
            + Ca_j * Ry_Ro * kiSRCa
            - kom * Ry_Ri
            + (Ca_j * Ca_j) * RI * koSRCa
        )
        J_SRCarel = ks * (Ca_sr - Ca_j) * Ry_Ro
        J_serca = (
            Vmax_SRCaP
            * ufl.elem_pow(Q10SRCaP, Qpow)
            * (
                -ufl.elem_pow(Ca_sr / Kmr, hillSRCaP)
                + ufl.elem_pow(Ca_i / Kmf, hillSRCaP)
            )
            / (
                1
                + ufl.elem_pow(Ca_sr / Kmr, hillSRCaP)
                + ufl.elem_pow(Ca_i / Kmf, hillSRCaP)
            )
        )
        J_SRleak = 5.348e-06 * Ca_sr - 5.348e-06 * Ca_j

        # Expressions for the Na Buffers component
        F_expressions[16] = -koff_na * Na_Bj + kon_na * (-Na_Bj + Bmax_Naj) * Na_j
        F_expressions[17] = kon_na * (-Na_Bsl + Bmax_Nasl) * Na_sl - koff_na * Na_Bsl

        # Expressions for the Cytosolic Ca Buffers component
        F_expressions[24] = kon_tncl * (Bmax_TnClow - Tn_CL) * Ca_i - koff_tncl * Tn_CL
        F_expressions[22] = (
            -koff_tnchca * Tn_CHc
            + kon_tnchca * (-Tn_CHc + Bmax_TnChigh - Tn_CHm) * Ca_i
        )
        F_expressions[23] = (
            Mgi * kon_tnchmg * (-Tn_CHc + Bmax_TnChigh - Tn_CHm) - koff_tnchmg * Tn_CHm
        )
        F_expressions[18] = kon_cam * (-CaM + Bmax_CaM) * Ca_i - koff_cam * CaM
        F_expressions[19] = (
            -koff_myoca * Myo_c + kon_myoca * (-Myo_c + Bmax_myosin - Myo_m) * Ca_i
        )
        F_expressions[20] = (
            Mgi * kon_myomg * (-Myo_c + Bmax_myosin - Myo_m) - koff_myomg * Myo_m
        )
        F_expressions[21] = kon_sr * (Bmax_SR - SRB) * Ca_i - koff_sr * SRB
        J_CaB_cytosol = (
            -koff_tnchca * Tn_CHc
            - koff_myoca * Myo_c
            + Mgi * kon_myomg * (-Myo_c + Bmax_myosin - Myo_m)
            + Mgi * kon_tnchmg * (-Tn_CHc + Bmax_TnChigh - Tn_CHm)
            - koff_tnchmg * Tn_CHm
            + kon_tncl * (Bmax_TnClow - Tn_CL) * Ca_i
            + kon_sr * (Bmax_SR - SRB) * Ca_i
            - koff_myomg * Myo_m
            + kon_cam * (-CaM + Bmax_CaM) * Ca_i
            - koff_cam * CaM
            - koff_tncl * Tn_CL
            + kon_myoca * (-Myo_c + Bmax_myosin - Myo_m) * Ca_i
            + kon_tnchca * (-Tn_CHc + Bmax_TnChigh - Tn_CHm) * Ca_i
            - koff_sr * SRB
        )

        # Expressions for the Junctional and SL Ca Buffers component
        Bmax_SLlowsl = Bmax_SLlowsl0 * Vmyo / Vsl
        Bmax_SLlowj = Bmax_SLlowj0 * Vmyo / Vjunc
        Bmax_SLhighsl = Bmax_SLhighsl0 * Vmyo / Vsl
        Bmax_SLhighj = Bmax_SLhighj0 * Vmyo / Vjunc
        F_expressions[27] = kon_sll * (Bmax_SLlowj - SLL_j) * Ca_j - koff_sll * SLL_j
        F_expressions[28] = (
            kon_sll * (-SLL_sl + Bmax_SLlowsl) * Ca_sl - koff_sll * SLL_sl
        )
        F_expressions[25] = kon_slh * (Bmax_SLhighj - SLH_j) * Ca_j - koff_slh * SLH_j
        F_expressions[26] = (
            kon_slh * (-SLH_sl + Bmax_SLhighsl) * Ca_sl - koff_slh * SLH_sl
        )
        J_CaB_junction = (
            kon_slh * (Bmax_SLhighj - SLH_j) * Ca_j
            + kon_sll * (Bmax_SLlowj - SLL_j) * Ca_j
            - koff_slh * SLH_j
            - koff_sll * SLL_j
        )
        J_CaB_sl = (
            kon_sll * (-SLL_sl + Bmax_SLlowsl) * Ca_sl
            + kon_slh * (-SLH_sl + Bmax_SLhighsl) * Ca_sl
            - koff_sll * SLL_sl
            - koff_slh * SLH_sl
        )

        # Expressions for the SR Ca Concentrations component
        Bmax_Csqn = Bmax_Csqn0 * Vmyo / Vsr
        F_expressions[30] = (
            -koff_csqn * Csqn_b + kon_csqn * (Bmax_Csqn - Csqn_b) * Ca_sr
        )
        F_expressions[29] = (
            -kon_csqn * (Bmax_Csqn - Csqn_b) * Ca_sr
            - J_SRleak * Vmyo / Vsr
            + koff_csqn * Csqn_b
            - J_SRCarel
            + J_serca
        )

        # Expressions for the Na Concentrations component
        I_Na_tot_junc = (
            3 * I_nak_junc + 3 * I_ncx_junc + I_CaNa_junc + I_Na_junc + I_nabk_junc
        )
        I_Na_tot_sl = I_Na_sl + I_nabk_sl + 3 * I_nak_sl + I_CaNa_sl + 3 * I_ncx_sl
        F_expressions[32] = (
            -Cmem * I_Na_tot_junc / (Frdy * Vjunc)
            + J_na_juncsl * (Na_sl - Na_j) / Vjunc
            - F_expressions[16]
        )
        F_expressions[33] = (
            -F_expressions[17]
            + J_na_slmyo * (-Na_sl + Na_i) / Vsl
            + J_na_juncsl * (-Na_sl + Na_j) / Vsl
            - Cmem * I_Na_tot_sl / (Frdy * Vsl)
        )
        F_expressions[31] = J_na_slmyo * (Na_sl - Na_i) / Vmyo

        # Expressions for the K Concentration component
        F_expressions[34] = backend.Constant(0.0)

        # Expressions for the Ca Concentrations component
        I_Ca_tot_junc = I_pca_junc + I_cabk_junc + I_Ca_junc - 2 * I_ncx_junc
        I_Ca_tot_sl = -2 * I_ncx_sl + I_pca_sl + I_cabk_sl + I_Ca_sl
        F_expressions[36] = (
            J_ca_juncsl * (Ca_sl - Ca_j) / Vjunc
            + J_SRCarel * Vsr / Vjunc
            - J_CaB_junction
            - Cmem * I_Ca_tot_junc / (2 * Frdy * Vjunc)
            + J_SRleak * Vmyo / Vjunc
        )
        F_expressions[37] = (
            -J_CaB_sl
            + J_ca_juncsl * (Ca_j - Ca_sl) / Vsl
            - Cmem * I_Ca_tot_sl / (2 * Frdy * Vsl)
            + J_ca_slmyo * (Ca_i - Ca_sl) / Vsl
        )
        F_expressions[35] = (
            -J_CaB_cytosol + J_ca_slmyo * (Ca_sl - Ca_i) / Vmyo - J_serca * Vsr / Vmyo
        )

        # Return results
        return dolfin.as_vector(F_expressions)

    def num_states(self):
        return 38

    def __str__(self):
        return "Grandi_pasqualini_bers_2010 cardiac cell model"



"""
These solvers solve the (pure) bidomain equations on the form: find
the transmembrane potential :math:`v = v(x, t)` and the extracellular
potential :math:`u = u(x, t)` such that

.. math::

   v_t - \mathrm{div} ( G_i v + G_i u) = I_s

   \mathrm{div} (G_i v + (G_i + G_e) u) = I_a

where the subscript :math:`t` denotes the time derivative; :math:`G_x`
denotes a weighted gradient: :math:`G_x = M_x \mathrm{grad}(v)` for
:math:`x \in \{i, e\}`, where :math:`M_i` and :math:`M_e` are the
intracellular and extracellular cardiac conductivity tensors,
respectively; :math:`I_s` and :math:`I_a` are prescribed input. In
addition, initial conditions are given for :math:`v`:

.. math::

   v(x, 0) = v_0

Finally, boundary conditions must be prescribed. For now, this solver
assumes pure homogeneous Neumann boundary conditions for :math:`v` and
:math:`u` and enforces the additional average value zero constraint
for u.

"""

# Copyright (C) 2013 Marie E. Rognes (meg@simula.no)
# Use and modify at will
# Last changed: 2013-04-18

__all__ = ["BasicBidomainSolver", "BidomainSolver"]

import ufl
import dolfin
from cbcbeat.dolfinimport import backend, has_dolfin_adjoint
from cbcbeat.markerwisefield import rhs_with_markerwise_field
from cbcbeat.utils import end_of_time, annotate_kwargs
from ufl.log import info, debug, error, warning


class BasicBidomainSolver(object):
    """This solver is based on a theta-scheme discretization in time
    and CG_1 x CG_1 (x R) elements in space.

    .. note::

       For the sake of simplicity and consistency with other solver
       objects, this solver operates on its solution fields (as state
       variables) directly internally. More precisely, solve (and
       step) calls will act by updating the internal solution
       fields. It implies that initial conditions can be set (and are
       intended to be set) by modifying the solution fields prior to
       simulation.

    *Arguments*
      mesh (:py:class:`dolfin.Mesh`)
        The spatial domain (mesh)

      time (:py:class:`dolfin.Constant` or None)
        A constant holding the current time. If None is given, time is
        created for you, initialized to zero.

      M_i (:py:class:`ufl.Expr`)
        The intracellular conductivity tensor (as an UFL expression)

      M_e (:py:class:`ufl.Expr`)
        The extracellular conductivity tensor (as an UFL expression)

      I_s (:py:class:`dict`, optional)
        A typically time-dependent external stimulus given as a dict,
        with domain markers as the key and a
        :py:class:`dolfin.Expression` as values. NB: it is assumed
        that the time dependence of I_s is encoded via the 'time'
        Constant.

      I_a (:py:class:`dolfin.Expression`, optional)
        A (typically time-dependent) external applied current

      v\_ (:py:class:`ufl.Expr`, optional)
        Initial condition for v. A new :py:class:`dolfin.Function`
        will be created if none is given.

      params (:py:class:`dolfin.Parameters`, optional)
        Solver parameters

    """

    def __init__(self, mesh, time, M_i, M_e, I_s=None, I_a=None, v_=None, params=None):
        # Check some input
        assert isinstance(mesh, dolfin.Mesh), (
            "Expecting mesh to be a Mesh instance, not %r" % mesh
        )
        assert (
            isinstance(time, backend.Constant) or time is None
        ), "Expecting time to be a Constant instance (or None)."
        assert (
            isinstance(params, dolfin.Parameters) or params is None
        ), "Expecting params to be a Parameters instance (or None)"

        self._nullspace_basis = None

        # Store input
        self._mesh = mesh
        self._time = time
        self._M_i = M_i
        self._M_e = M_e
        self._I_s = I_s
        self._I_a = I_a

        # Initialize and update parameters if given
        self.parameters = self.default_parameters()
        if params is not None:
            self.parameters.update(params)

        # Set-up function spaces
        k = self.parameters["polynomial_degree"]
        Ve = dolfin.FiniteElement("CG", self._mesh.ufl_cell(), k)
        V = dolfin.FunctionSpace(self._mesh, "CG", k)
        Ue = dolfin.FiniteElement("CG", self._mesh.ufl_cell(), k)
        dolfin.FunctionSpace(self._mesh, "CG", k)

        use_R = self.parameters["use_avg_u_constraint"]
        if use_R:
            Re = dolfin.FiniteElement("R", self._mesh.ufl_cell(), 0)
            dolfin.FunctionSpace(self._mesh, "R", 0)
            self.VUR = dolfin.FunctionSpace(mesh, dolfin.MixedElement((Ve, Ue, Re)))
        else:
            self.VUR = dolfin.FunctionSpace(mesh, dolfin.MixedElement((Ve, Ue)))

        self.V = V

        # Set-up solution fields:
        if v_ is None:
            self.merger = backend.FunctionAssigner(V, self.VUR.sub(0))
            self.v_ = backend.Function(V, name="v_")
        else:
            debug("Experimental: v_ shipped from elsewhere.")
            self.merger = None
            self.v_ = v_
        self.vur = backend.Function(self.VUR, name="vur")

        # Figure out whether we should annotate or not
        self._annotate_kwargs = annotate_kwargs(self.parameters)

    @property
    def time(self):
        "The internal time of the solver."
        return self._time

    def solution_fields(self):
        """
        Return tuple of previous and current solution objects.

        Modifying these will modify the solution objects of the solver
        and thus provides a way for setting initial conditions for
        instance.

        *Returns*
          (previous v, current vur) (:py:class:`tuple` of :py:class:`dolfin.Function`)
        """
        return (self.v_, self.vur)

    def solve(self, interval, dt=None):
        """
        Solve the discretization on a given time interval (t0, t1)
        with a given timestep dt and return generator for a tuple of
        the interval and the current solution.

        *Arguments*
          interval (:py:class:`tuple`)
            The time interval for the solve given by (t0, t1)
          dt (int, optional)
            The timestep for the solve. Defaults to length of interval

        *Returns*
          (timestep, solution_fields) via (:py:class:`genexpr`)

        *Example of usage*::

          # Create generator
          solutions = solver.solve((0.0, 1.0), 0.1)

          # Iterate over generator (computes solutions as you go)
          for (interval, solution_fields) in solutions:
            (t0, t1) = interval
            v_, vur = solution_fields
            # do something with the solutions
        """
        dolfin.Timer("PDE step")

        # Initial set-up
        # Solve on entire interval if no interval is given.
        (T0, T) = interval
        if dt is None:
            dt = T - T0
        t0 = T0
        t1 = T0 + dt

        # Step through time steps until at end time
        while True:
            info("Solving on t = (%g, %g)" % (t0, t1))
            self.step((t0, t1))

            # Yield solutions
            yield (t0, t1), self.solution_fields()

            # Break if this is the last step
            if end_of_time(T, t0, t1, dt):
                break

            # If not: update members and move to next time
            # Subfunction assignment would be good here.
            if isinstance(self.v_, backend.Function):
                self.merger.assign(self.v_, self.vur.sub(0))
            else:
                debug("Assuming that v_ is updated elsewhere. Experimental.")
            t0 = t1
            t1 = t0 + dt

    def step(self, interval):
        """
        Solve on the given time interval (t0, t1).

        *Arguments*
          interval (:py:class:`tuple`)
            The time interval (t0, t1) for the step

        *Invariants*
          Assuming that v\_ is in the correct state for t0, gives
          self.vur in correct state at t1.
        """

        dolfin.Timer("PDE step")

        # Extract interval and thus time-step
        (t0, t1) = interval
        k_n = backend.Constant(t1 - t0)
        theta = self.parameters["theta"]

        # Extract conductivities
        M_i, M_e = self._M_i, self._M_e

        # Define variational formulation
        use_R = self.parameters["use_avg_u_constraint"]
        if use_R:
            (v, u, l) = dolfin.TrialFunctions(self.VUR)
            (w, q, lamda) = dolfin.TestFunctions(self.VUR)
        else:
            (v, u) = dolfin.TrialFunctions(self.VUR)
            (w, q) = dolfin.TestFunctions(self.VUR)

        Dt_v = (v - self.v_) / k_n
        v_mid = theta * v + (1.0 - theta) * self.v_

        # Set time
        t = t0 + theta * (t1 - t0)
        self.time.assign(t)

        # Define spatial integration domains:
        (dz, rhs) = rhs_with_markerwise_field(self._I_s, self._mesh, w)

        theta_parabolic = (
            ufl.inner(M_i * ufl.grad(v_mid), ufl.grad(w)) * dz()
            + ufl.inner(M_i * ufl.grad(u), ufl.grad(w)) * dz()
        )
        theta_elliptic = (
            ufl.inner(M_i * ufl.grad(v_mid), ufl.grad(q)) * dz()
            + ufl.inner((M_i + M_e) * ufl.grad(u), ufl.grad(q)) * dz()
        )
        G = Dt_v * w * dz() + theta_parabolic + theta_elliptic

        if use_R:
            G += (lamda * u + l * q) * dz()

        # Add applied current as source in elliptic equation if
        # applicable
        if self._I_a:
            G -= self._I_a * q * dz()

        # Add applied stimulus as source in parabolic equation if
        # applicable
        G -= rhs

        # Define variational problem
        a, L = dolfin.system(G)
        pde = backend.LinearVariationalProblem(a, L, self.vur)

        # Set-up solver
        solver = backend.LinearVariationalSolver(pde)
        solver.parameters.update(self.parameters["linear_variational_solver"])
        solver.solve()

    @staticmethod
    def default_parameters():
        """Initialize and return a set of default parameters

        *Returns*
          A set of parameters (:py:class:`dolfin.Parameters`)

        To inspect all the default parameters, do::

          info(BasicBidomainSolver.default_parameters(), True)
        """

        params = dolfin.Parameters("BasicBidomainSolver")
        params.add("enable_adjoint", True)
        params.add("theta", 0.5)
        params.add("polynomial_degree", 1)
        params.add("use_avg_u_constraint", True)

        params.add(backend.LinearVariationalSolver.default_parameters())
        return params


class BidomainSolver(BasicBidomainSolver):
    __doc__ = BasicBidomainSolver.__doc__

    def __init__(self, mesh, time, M_i, M_e, I_s=None, I_a=None, v_=None, params=None):
        # Call super-class
        BasicBidomainSolver.__init__(
            self, mesh, time, M_i, M_e, I_s=I_s, I_a=I_a, v_=v_, params=params
        )

        # Check consistency of parameters first
        if self.parameters["enable_adjoint"] and not has_dolfin_adjoint:
            warning(
                "'enable_adjoint' is set to True, but no " "dolfin_adjoint installed."
            )

        # Mark the timestep as unset
        self._timestep = None

    @property
    def linear_solver(self):
        """The linear solver (:py:class:`dolfin.LUSolver` or
        :py:class:`dolfin.PETScKrylovSolver`)."""
        return self._linear_solver

    def _create_linear_solver(self):
        "Helper function for creating linear solver based on parameters."
        solver_type = self.parameters["linear_solver_type"]

        if solver_type == "direct":
            solver = backend.LUSolver(self._lhs_matrix)
            solver.parameters.update(self.parameters["lu_solver"])
            update_routine = self._update_lu_solver

        elif solver_type == "iterative":
            # Initialize KrylovSolver with matrix
            alg = self.parameters["algorithm"]
            prec = self.parameters["preconditioner"]

            debug("Creating PETSCKrylovSolver with %s and %s" % (alg, prec))
            if prec == "fieldsplit":
                # Argh. DOLFIN won't let you construct a PETScKrylovSolver with fieldsplit. Sigh ..
                solver = backend.PETScKrylovSolver()
                # FIXME: work around DOLFIN bug #583. Just deleted this when fixed.
                solver.parameters.update({"convergence_norm_type": "preconditioned"})
                # solver.parameters["preconditioner"]["structure"] = "same"
                # MER this should be set by user, and is below
                solver.parameters.update(self.parameters["petsc_krylov_solver"])
                solver.set_operator(self._lhs_matrix)

                # Initialize the KSP directly:
                ksp = solver.ksp()
                ksp.setType(alg)
                ksp.pc.setType(prec)
                ksp.setOptionsPrefix(
                    "bidomain_"
                )  # it's really stupid, solver.set_options_prefix() doesn't work

                # Set various options (by default) for the fieldsplit
                # approach to solving the bidomain equations.

                # FIXME: This needs a try
                from petsc4py import PETSc

                # Patrick believes that the fieldsplit index sets
                # should already be set from the assembled matrix.

                # Now let's set some default options for the solver.
                opts = PETSc.Options("bidomain_")
                if "pc_fieldsplit_type" not in opts:
                    opts["pc_fieldsplit_type"] = "symmetric_multiplicative"
                if "fieldsplit_0_ksp_type" not in opts:
                    opts["fieldsplit_0_ksp_type"] = "preonly"
                if "fieldsplit_1_ksp_type" not in opts:
                    opts["fieldsplit_1_ksp_type"] = "preonly"
                if "fieldsplit_0_pc_type" not in opts:
                    opts["fieldsplit_0_pc_type"] = "hypre"
                if "fieldsplit_1_pc_type" not in opts:
                    opts["fieldsplit_1_pc_type"] = "hypre"

                ksp.setFromOptions()
                ksp.setUp()

            else:
                solver = backend.PETScKrylovSolver(alg, prec)
                solver.set_operator(self._lhs_matrix)
                # Still waiting for that bug fix:
                solver.parameters.update({"convergence_norm_type": "preconditioned"})
                solver.parameters.update(self.parameters["petsc_krylov_solver"])

            # Set nullspace if present. We happen to know that the
            # transpose nullspace is the same as the nullspace (easy
            # to prove from matrix structure).
            if self.parameters["use_avg_u_constraint"]:
                # Nothing to do, no null space
                pass

            else:
                # If dolfin-adjoint is enabled and installled: set the solver nullspace
                if has_dolfin_adjoint:
                    solver.set_nullspace(self.nullspace)
                    solver.set_transpose_nullspace(self.nullspace)
                # Otherwise, set the nullspace in the operator
                # directly.
                else:
                    A = dolfin.as_backend_type(self._lhs_matrix)
                    A.set_nullspace(self.nullspace)

            update_routine = self._update_krylov_solver
        else:
            error("Unknown linear_solver_type given: %s" % solver_type)

        return (solver, update_routine)

    @property
    def nullspace(self):
        if self._nullspace_basis is None:
            null_vector = dolfin.Vector(self.vur.vector())
            self.VUR.sub(1).dofmap().set(null_vector, 1.0)
            null_vector *= 1.0 / null_vector.norm("l2")
            self._nullspace_basis = dolfin.VectorSpaceBasis([null_vector])
        return self._nullspace_basis

    @staticmethod
    def default_parameters():
        """Initialize and return a set of default parameters

        *Returns*
          A set of parameters (:py:class:`dolfin.Parameters`)

        To inspect all the default parameters, do::

          info(BidomainSolver.default_parameters(), True)
        """
        params = dolfin.Parameters("BidomainSolver")
        params.add("enable_adjoint", True)
        params.add("theta", 0.5)
        params.add("polynomial_degree", 1)

        # Set default solver type to be iterative
        params.add("linear_solver_type", "iterative")
        params.add("use_avg_u_constraint", False)

        # Set default iterative solver choices (used if iterative
        # solver is invoked)
        params.add("algorithm", "cg")
        params.add("preconditioner", "petsc_amg")
        # params.add("preconditioner", "fieldsplit") # This seg faults

        # Add default parameters from both LU and Krylov solvers
        params.add(backend.LUSolver.default_parameters())
        petsc_params = backend.PETScKrylovSolver.default_parameters()
        # FIXME: work around DOLFIN bug #583. Just deleted this when fixed.
        petsc_params.update({"convergence_norm_type": "preconditioned"})
        params.add(petsc_params)

        # Customize default parameters for PETScKrylovSolver
        # params["petsc_krylov_solver"]["preconditioner"]["structure"] = "same"

        return params

    def variational_forms(self, k_n):
        """Create the variational forms corresponding to the given
        discretization of the given system of equations.

        *Arguments*
          k_n (:py:class:`ufl.Expr` or float)
            The time step

        *Returns*
          (lhs, rhs) (:py:class:`tuple` of :py:class:`ufl.Form`)

        """

        # Extract theta parameter and conductivities
        theta = self.parameters["theta"]
        M_i = self._M_i
        M_e = self._M_e

        # Define variational formulation
        use_R = self.parameters["use_avg_u_constraint"]
        if use_R:
            (v, u, l) = dolfin.TrialFunctions(self.VUR)
            (w, q, lamda) = dolfin.TestFunctions(self.VUR)
        else:
            (v, u) = dolfin.TrialFunctions(self.VUR)
            (w, q) = dolfin.TestFunctions(self.VUR)

        # Set-up measure and rhs from stimulus
        (dz, rhs) = rhs_with_markerwise_field(self._I_s, self._mesh, w)

        # Set-up variational problem
        Dt_v_k_n = v - self.v_
        v_mid = theta * v + (1.0 - theta) * self.v_
        theta_parabolic = (
            ufl.inner(M_i * ufl.grad(v_mid), ufl.grad(w)) * dz()
            + ufl.inner(M_i * ufl.grad(u), ufl.grad(w)) * dz()
        )
        theta_elliptic = (
            ufl.inner(M_i * ufl.grad(v_mid), ufl.grad(q)) * dz()
            + ufl.inner((M_i + M_e) * ufl.grad(u), ufl.grad(q)) * dz()
        )

        G = (
            Dt_v_k_n * w * dz()
            + k_n * theta_parabolic
            + k_n * theta_elliptic
            - k_n * rhs
        )

        if use_R:
            G += k_n * (lamda * u + l * q) * dz()

        # Add applied current as source in elliptic equation if
        # applicable
        if self._I_a:
            G -= k_n * self._I_a * q * dz()

        (a, L) = dolfin.system(G)
        return (a, L)

    def step(self, interval):
        """
        Solve on the given time step (t0, t1).

        *Arguments*
          interval (:py:class:`tuple`)
            The time interval (t0, t1) for the step

        *Invariants*
          Assuming that v\_ is in the correct state for t0, gives
          self.vur in correct state at t1.
        """

        dolfin.Timer("PDE step")
        self.parameters["linear_solver_type"]

        # Extract interval and thus time-step
        (t0, t1) = interval
        dt = t1 - t0
        theta = self.parameters["theta"]
        t = t0 + theta * dt
        self.time.assign(t)

        # Update matrix and linear solvers etc as needed
        if self._timestep is None:
            self._timestep = backend.Constant(dt)
            (self._lhs, self._rhs) = self.variational_forms(self._timestep)

            # Preassemble left-hand side and initialize right-hand side vector
            debug("Preassembling bidomain matrix (and initializing vector)")
            self._lhs_matrix = backend.assemble(self._lhs, **self._annotate_kwargs)
            self._rhs_vector = dolfin.Vector(
                self._mesh.mpi_comm(), self._lhs_matrix.size(0)
            )
            self._lhs_matrix.init_vector(self._rhs_vector, 0)

            # Create linear solver (based on parameter choices)
            self._linear_solver, self._update_solver = self._create_linear_solver()
        else:
            timestep_unchanged = abs(dt - float(self._timestep)) < 1.0e-12
            self._update_solver(timestep_unchanged, dt)

        # Assemble right-hand-side
        backend.assemble(self._rhs, tensor=self._rhs_vector, **self._annotate_kwargs)

        # Solve problem
        self.linear_solver.solve(
            self.vur.vector(), self._rhs_vector, **self._annotate_kwargs
        )

    def _update_lu_solver(self, timestep_unchanged, dt):
        """Helper function for updating an LUSolver depending on
        whether timestep has changed."""

        # Update reuse of factorization parameter in accordance with
        # changes in timestep
        if timestep_unchanged:
            debug("Timestep is unchanged, reusing LU factorization")
        else:
            debug("Timestep has changed, updating LU factorization")
            if has_dolfin_adjoint and self.parameters["enable_adjoint"]:
                raise ValueError(
                    "dolfin-adjoint doesn't support changing timestep (yet)"
                )

            # Update stored timestep
            # FIXME: dolfin_adjoint still can't annotate constant assignment.
            self._timestep.assign(backend.Constant(dt))  # , annotate=annotate)

            # Reassemble matrix
            backend.assemble(
                self._lhs, tensor=self._lhs_matrix, **self._annotate_kwargs
            )

            (self._linear_solver, dummy) = self._create_linear_solver()

    def _update_krylov_solver(self, timestep_unchanged, dt):
        """Helper function for updating a KrylovSolver depending on
        whether timestep has changed."""

        # Update reuse of preconditioner parameter in accordance with
        # changes in timestep
        if timestep_unchanged:
            debug("Timestep is unchanged, reusing preconditioner")
        else:
            debug("Timestep has changed, updating preconditioner")
            if has_dolfin_adjoint and self.parameters["enable_adjoint"]:
                raise ValueError(
                    "dolfin-adjoint doesn't support changing timestep (yet)"
                )

            # Update stored timestep
            self._timestep.assign(backend.Constant(dt))  # , annotate=annotate)

            # Reassemble matrix
            backend.assemble(
                self._lhs, tensor=self._lhs_matrix, **self._annotate_kwargs
            )

            # Make new Krylov solver
            (self._linear_solver, dummy) = self._create_linear_solver()

        # Set nonzero initial guess if it indeed is nonzero
        if self.vur.vector().norm("l2") > 1.0e-12:
            debug("Initial guess is non-zero.")
            self.linear_solver.parameters["nonzero_initial_guess"] = True



# Copyright (C) 2014 Marie E. Rognes (meg@simula.no)
# Use and modify at will
# Last changed: 2016-10-19

__all__ = ["Markerwise", "handle_markerwise", "rhs_with_markerwise_field"]

from dolfin import dx, Measure
from ufl.log import error


def handle_markerwise(g, classtype):
    # Handle stimulus
    if (
        g is None
        or isinstance(g, classtype)
        or isinstance(g, Markerwise)
        or isinstance(g, object)
    ):
        return g
    else:
        msg = "Expecting stimulus to be a %s or Markerwise, not %r " % (
            str(classtype),
            g,
        )
        error(msg)


def rhs_with_markerwise_field(g, mesh, v):
    if g is None:
        dz = dx
        rhs = 0.0
    elif isinstance(g, Markerwise):
        markers = g.markers()
        dz = Measure("dx", domain=mesh, subdomain_data=markers)
        rhs = sum([g * v * dz(i) for (i, g) in zip(g.keys(), g.values())])
    else:
        dz = dx
        rhs = g * v * dz()
    return (dz, rhs)


class Markerwise(object):
    """A container class representing an object defined by a number of
    objects combined with a mesh function defining mesh domains and a
    map between the these.

    *Arguments*
      objects (tuple)
        the different objects
      keys (tuple of ints)
        a map from the objects to the domains marked in markers
      markers (:py:class:`dolfin.MeshFunction`)
        a mesh function mapping which domains the mesh consist of

    *Example of usage*

    Given (g0, g1), (2, 5) and markers, let

      g = g0 on domains marked by 2 in markers
      g = g1 on domains marked by 5 in markers

    letting::

      g = Markerwise((g0, g1), (2, 5), markers)

    """

    def __init__(self, objects, keys, markers):
        "Create Markerwise from given input."

        # Check input
        assert len(objects) == len(
            keys
        ), "Expecting the number of objects to equal the number of keys"

        # Store attributes:
        self._objects = dict(zip(keys, objects))
        self._markers = markers

    def values(self):
        "The objects"
        return self._objects.values()

    def keys(self):
        "The keys or domain numbers"
        return self._objects.keys()

    def markers(self):
        "The markers"
        return self._markers

    def __getitem__(self, key):
        "The objects"
        return self._objects[key]



# Copyright (C) 2012 Johan Hake (hake.dev@gmail.com)
# Use and modify at will
# Last changed: 2014-09-12

try:
    import gotran
except Exception as e:
    print("Gotran not installed. Not possible to convert gotran model to cellmodel.")
    raise e

# Gotran imports
from gotran.model import load_ode
from gotran.model.ode import ODE
from modelparameters.logger import error as gotran_error
from modelparameters.utils import check_arg

from gotran.common.options import parameters as gotran_parameters
from gotran.codegeneration.algorithmcomponents import componentwise_derivative
from gotran.codegeneration.codecomponent import CodeComponent

from gotran.codegeneration.codegenerators import DOLFINCodeGenerator
from ufl.log import error

_class_template = """
\"\"\"This module contains a {ModelName} cardiac cell model

The module was autogenerated from a gotran ode file
\"\"\"
from __future__ import division
from collections import OrderedDict
import ufl

from cbcbeat.dolfinimport import *
from cbcbeat.cellmodels import CardiacCellModel

class {ModelName}(CardiacCellModel):
    def __init__(self, params=None, init_conditions=None):
        \"\"\"
        Create cardiac cell model

        *Arguments*
         params (dict, :py:class:`dolfin.Mesh`, optional)
           optional model parameters
         init_conditions (dict, :py:class:`dolfin.Mesh`, optional)
           optional initial conditions
        \"\"\"
        CardiacCellModel.__init__(self, params, init_conditions)

    @staticmethod
    def default_parameters():
        \"Set-up and return default parameters.\"
{default_parameters}
        return params

    @staticmethod
    def default_initial_conditions():
        \"Set-up and return default initial conditions.\"
{initial_conditions}
        return ic

    def _I(self, v, s, time):
        \"\"\"
        Original gotran transmembrane current dV/dt
        \"\"\"
{I_body}

    def I(self, v, s, time=None):
        \"\"\"
        Transmembrane current

           I = -dV/dt

        \"\"\"
        return -self._I(v, s, time)

    def F(self, v, s, time=None):
        \"\"\"
        Right hand side for ODE system
        \"\"\"
{F_body}

    def num_states(self):
        return {num_states}

    def __str__(self):
        return '{ModelName} cardiac cell model'
"""

_class_form = dict(
    ModelName="NOT_IMPLEMENTED",
    default_parameters="NOT_IMPLEMENTED",
    I_body="NOT_IMPLEMENTED",
    F_body="NOT_IMPLEMENTED",
    num_states="NOT_IMPLEMENTED",
    initial_conditions="NOT_IMPLEMENTED",
)


__all__ = ["CellModelGenerator"]


class CellModelGenerator(DOLFINCodeGenerator):
    """
    Convert a Gotran model to a cbcbeat compatible cell model
    """

    def __init__(self, ode, membrane_potential):
        # Init base class
        super(CellModelGenerator, self).__init__()

        check_arg(ode, ODE, 0)
        check_arg(membrane_potential, str, 1)

        assert not ode.is_dae

        # Capitalize first letter of name
        name = ode.name
        self.name = (
            name
            if name[0].isupper()
            else name[0].upper() + (name[1:] if len(name) > 1 else "")
        )

        # Set cbcbeat compatible gotran code generation parameters
        generation_params = gotran_parameters.generation.copy()
        generation_params.code.default_arguments = "stp"
        generation_params.code.time.name = "time"

        generation_params.code.array.index_format = "[]"
        generation_params.code.array.index_offset = 0

        generation_params.code.parameters.representation = "named"
        generation_params.code.parameters.array_name = "parameters"
        generation_params.code.states.representation = "named"
        generation_params.code.states.array_name = "states"
        generation_params.code.body.representation = "named"
        generation_params.code.body.use_cse = False

        if ode.num_full_states < 2:
            gotran_error("expected the ODE to have at least more than 1 state")

        # Check that ode model has the membrane potential state name
        if membrane_potential not in ode.present_ode_objects:
            gotran_error(
                "Cannot find the membrane potential. ODE does not "
                "contain a state with name '{0}'".format(membrane_potential)
            )

        state = ode.present_ode_objects[membrane_potential]
        if not isinstance(state, gotran.model.State):
            gotran_error(
                "Cannot find the membrane potential. ODE does not "
                "contain a state with name '{0}'".format(membrane_potential)
            )

        # The name of the membrane potential
        self.V_name = state.name

        # Get the I and F expressions
        I_ind = [
            ind
            for ind, expr in enumerate(ode.state_expressions)
            if expr.state.name == self.V_name
        ][0]

        # Create gotran code component for I(s,t) (dV/dt)
        I_comp = componentwise_derivative(
            ode, I_ind, generation_params.code, result_name="current"
        )

        # Create gotran code component for F(s,t) (dS/dt)
        F_inds = list(range(ode.num_full_states))
        F_inds.remove(I_ind)
        F_comp = componentwise_derivative(
            ode, F_inds, generation_params.code, result_name="F_expressions"
        )

        # Create the class form and start fill it
        self._class_form = _class_form.copy()
        self._class_form["I_body"] = self.function_code(
            I_comp, indent=2, include_signature=False
        )
        self._class_form["F_body"] = self.function_code(
            F_comp, indent=2, include_signature=False
        )
        self._class_form["num_states"] = ode.num_full_states - 1
        self._class_form["ModelName"] = self.name
        self._class_form["default_parameters"] = self.default_parameters_body(ode)
        self._class_form["initial_conditions"] = self.initial_conditions_body(ode)

    def _init_arguments(self, comp, default_arguments=None):
        check_arg(comp, CodeComponent)
        params = self.params.code
        default_arguments = default_arguments or params.default_arguments

        # Check if comp defines used_states if not use the root components
        # full_states attribute
        # FIXME: No need for full_states here...
        states = [state.name for state in comp.root.full_states]
        used_parameters = comp.used_parameters

        num_states = comp.root.num_full_states

        # Start building body
        body_lines = []
        body_lines.append("time = time if time else Constant(0.0)")
        body_lines.append("")
        body_lines.append("# Assign states")
        if self.V_name != "v":
            body_lines.append("{0} = v".format(self.V_name))

        state_names = [state for state in states if state != self.V_name]
        if len(states) == 1:
            if states[0] != "s":
                body_lines.append("{0} = s".format(states[0]))
        else:
            body_lines.append("assert(len(s) == {0})".format(num_states - 1))
            body_lines.append(", ".join(state_names) + " = s")
        body_lines.append("")

        body_lines.append("# Assign parameters")
        for param in used_parameters:
            body_lines.append(
                "{0} = self._parameters[" '"{1}"]'.format(param.name, param.name)
            )

        # If initilizing results
        if comp.results:
            body_lines.append("")
            body_lines.append("# Init return args")

        for result_name in comp.results:
            shape = comp.shapes[result_name]
            if len(shape) > 1:
                error("expected only result expression with rank 1")

            body_lines.append("{0} = [ufl.zero()]*{1}".format(result_name, shape[0]))

        return body_lines

    def generate(self):
        """
        Return a beat cell model file as a str
        """
        return _class_template.format(**self._class_form)

    def default_parameters_body(self, ode):
        """
        Generate code for the default parameter bod
        """
        if ode.num_parameters > 0:
            params = ode.parameters[:]

            param = params.pop(0)

            body_lines = [
                'params = OrderedDict([("{}", {}),'.format(param.name, param.init)
            ]
            body_lines.extend(
                '                      ("{}", {}),'.format(
                    param.name,
                    param.init
                    if isinstance(param.init, (float, int))
                    else param.init[0],
                )
                for param in params
            )
            body_lines[-1] = body_lines[-1][0:-1] + "])"
        else:
            body_lines = ["params = OrderedDict()"]

        return "\n".join(self.indent_and_split_lines(body_lines, 2))

    def initial_conditions_body(self, ode):
        """
        Generate code for the ic body
        """

        # First get ic for v
        v_init = ode.present_ode_objects[self.V_name].init
        s_init, s_names = zip(
            *[
                (state.init, state.name)
                for state in ode.full_states
                if state.name != self.V_name
            ]
        )
        body_lines = ['ic = OrderedDict([("{}", {}),'.format(self.V_name, v_init)]
        body_lines.extend(
            '                  ("{}", {}),'.format(name, value)
            for name, value in zip(s_names, s_init)
        )
        body_lines[-1] = body_lines[-1][0:-1] + "])"

        return "\n".join(self.indent_and_split_lines(body_lines, 2))


def gotran2beat(filename, params):
    """
    Create a beat cell model from a gotran model
    """

    # Load Gotran model
    ode = load_ode(filename)

    # Create a Beat Cell model code generator
    cell_gen = CellModelGenerator(ode, params.membrane_potential)

    output = params.output

    if output:
        if not (len(output) > 3 and output[-3:] == ".py"):
            output += ".py"
    else:
        output = filename.replace(".ode", "") + ".py"

    f = open(output, "w")

    f.write(cell_gen.generate())


def main():
    import sys
    import os
    from modelparameters.parameterdict import ParameterDict, Param

    params = ParameterDict(
        output=Param("", description="Specify the basename of the output file"),
        membrane_potential=Param(
            "V", description="The name of the " "membrane potential state."
        ),
    )
    params.parse_args(usage="usage: %prog FILE [options]")  # sys.argv[2:])

    if len(sys.argv) < 2:
        raise RuntimeError("Expected a single gotran file argument")

    if not os.path.isfile(sys.argv[1]):
        raise IOError("Expected the argument to be a file")

    file_name = sys.argv[1]
    gotran2beat(file_name, params)


if __name__ == "__main__":
    raise SystemExit(main())



"""This module contains a container class for cardiac models:
:py:class:`~cbcbeat.cardiacmodels.CardiacModel`.  This class
should be instantiated for setting up specific cardiac simulation
scenarios.
"""

# Copyright (C) 2012 Marie E. Rognes (meg@simula.no)
# Use and modify at will
# Last changed: 2016-04-21

__all__ = ["CardiacModel"]

import dolfin
from cbcbeat.dolfinimport import backend
from cbcbeat.markerwisefield import Markerwise, handle_markerwise
from cbcbeat.cellmodels import CardiacCellModel
from ufl.log import error

# ------------------------------------------------------------------------------
# Cardiac models
# ------------------------------------------------------------------------------


class CardiacModel(object):
    """
    A container class for cardiac models. Objects of this class
    represent a specific cardiac simulation set-up and should provide

    * A computational domain
    * A cardiac cell model
    * Intra-cellular and extra-cellular conductivities
    * Various forms of stimulus (optional).

    This container class is designed for use with the splitting
    solvers (:py:mod:`cbcbeat.splittingsolver`), see their
    documentation for more information on how the attributes are
    interpreted in that context.

    *Arguments*
      domain (:py:class:`dolfin.Mesh`)
        the computational domain in space
      time (:py:class:`dolfin.Constant` or None )
        A constant holding the current time.
      M_i (:py:class:`ufl.Expr`)
        the intra-cellular conductivity as an ufl Expression
      M_e (:py:class:`ufl.Expr`)
        the extra-cellular conductivity as an ufl Expression
      cell_models (:py:class:`~cbcbeat.cellmodels.cardiaccellmodel.CardiacCellModel`)
        a cell model or a dict with cell models associated with a cell model domain
      stimulus (:py:class:`dict`, optional)
        A typically time-dependent external stimulus given as a dict,
        with domain markers as the key and a
        :py:class:`dolfin.Expression` as values. NB: it is assumed
        that the time dependence of I_s is encoded via the 'time'
        Constant.
      applied_current (:py:class:`ufl.Expr`, optional)
        an applied current as an ufl Expression

    """

    def __init__(
        self, domain, time, M_i, M_e, cell_models, stimulus=None, applied_current=None
    ):
        "Create CardiacModel from given input."

        self._handle_input(
            domain, time, M_i, M_e, cell_models, stimulus, applied_current
        )

    def _handle_input(
        self, domain, time, M_i, M_e, cell_models, stimulus=None, applied_current=None
    ):
        # Check input and store attributes
        msg = "Expecting domain to be a Mesh instance, not %r" % domain
        assert isinstance(domain, dolfin.Mesh), msg
        self._domain = domain

        msg = "Expecting time to be a Constant instance, not %r." % time
        assert isinstance(time, backend.Constant) or time is None, msg
        self._time = time

        self._intracellular_conductivity = M_i
        self._extracellular_conductivity = M_e

        # Handle cell_models
        self._cell_models = handle_markerwise(cell_models, CardiacCellModel)
        if isinstance(self._cell_models, Markerwise):
            msg = "Different cell_models are currently not supported."
            error(msg)

        # Handle stimulus
        self._stimulus = handle_markerwise(
            stimulus, dolfin.cpp.function.GenericFunction
        )

        # Handle applied current
        ac = applied_current
        self._applied_current = handle_markerwise(
            ac, dolfin.cpp.function.GenericFunction
        )

    def applied_current(self):
        "An applied current: used as a source in the elliptic bidomain equation"
        return self._applied_current

    def stimulus(self):
        "A stimulus: used as a source in the parabolic bidomain equation"
        return self._stimulus

    def conductivities(self):
        """Return the intracellular and extracellular conductivities
        as a tuple of UFL Expressions.

        *Returns*
        (M_i, M_e) (:py:class:`tuple` of :py:class:`ufl.Expr`)
        """
        return (self.intracellular_conductivity(), self.extracellular_conductivity())

    def intracellular_conductivity(self):
        "The intracellular conductivity (:py:class:`ufl.Expr`)."
        return self._intracellular_conductivity

    def extracellular_conductivity(self):
        "The intracellular conductivity (:py:class:`ufl.Expr`)."
        return self._extracellular_conductivity

    def time(self):
        "The current time (:py:class:`dolfin.Constant` or None)."
        return self._time

    def domain(self):
        "The spatial domain (:py:class:`dolfin.Mesh`)."
        return self._domain

    def cell_models(self):
        "Return the cell models"
        return self._cell_models



"""This module contains a Fenton_karma_1998_BR_altered cardiac cell model

The module was autogenerated from a gotran ode file
"""
from __future__ import division
from collections import OrderedDict
import ufl

import dolfin
from cbcbeat.dolfinimport import backend
from cbcbeat.cellmodels.cardiaccellmodel import CardiacCellModel


class Fenton_karma_1998_BR_altered(CardiacCellModel):
    def __init__(self, params=None, init_conditions=None):
        """
        Create cardiac cell model

        *Arguments*
         params (dict, :py:class:`dolfin.Mesh`, optional)
           optional model parameters
         init_conditions (dict, :py:class:`dolfin.Mesh`, optional)
           optional initial conditions
        """
        CardiacCellModel.__init__(self, params, init_conditions)

    @staticmethod
    def default_parameters():
        "Set-up and return default parameters."
        params = OrderedDict(
            [
                ("u_c", 0.13),
                ("u_v", 0.04),
                ("g_fi_max", 4),
                ("tau_v1_minus", 1250),
                ("tau_v2_minus", 19.6),
                ("tau_v_plus", 3.33),
                ("tau_0", 12.5),
                ("tau_r", 33.33),
                ("k", 10),
                ("tau_si", 29),
                ("u_csi", 0.85),
                ("tau_w_minus", 41),
                ("tau_w_plus", 870),
                ("IstimAmplitude", -0.2),
                ("IstimEnd", 50000),
                ("IstimPeriod", 1000),
                ("IstimPulseDuration", 1),
                ("IstimStart", 10),
                ("Cm", 1),
                ("V_0", -85),
                ("V_fi", 15),
            ]
        )
        return params

    @staticmethod
    def default_initial_conditions():
        "Set-up and return default initial conditions."
        ic = OrderedDict([("V", -85), ("v", 1), ("w", 1)])
        return ic

    def _I(self, v, s, time):
        """
        Original gotran transmembrane current dV/dt
        """
        time = time if time else backend.Constant(0.0)

        # Assign states
        V = v
        assert len(s) == 2
        v, w = s

        # Assign parameters
        u_c = self._parameters["u_c"]
        g_fi_max = self._parameters["g_fi_max"]
        tau_0 = self._parameters["tau_0"]
        tau_r = self._parameters["tau_r"]
        k = self._parameters["k"]
        tau_si = self._parameters["tau_si"]
        u_csi = self._parameters["u_csi"]
        Cm = self._parameters["Cm"]
        V_0 = self._parameters["V_0"]
        V_fi = self._parameters["V_fi"]

        # Init return args
        current = [ufl.zero()] * 1

        # Expressions for the p component
        p = ufl.conditional(ufl.lt((-V_0 + V) / (V_fi - V_0), u_c), 0, 1)

        # Expressions for the Fast inward current component
        tau_d = Cm / g_fi_max
        J_fi = (
            -(1 - (-V_0 + V) / (V_fi - V_0))
            * (-u_c + (-V_0 + V) / (V_fi - V_0))
            * p
            * v
            / tau_d
        )

        # Expressions for the Slow outward current component
        J_so = p / tau_r + (1 - p) * (-V_0 + V) / (tau_0 * (V_fi - V_0))

        # Expressions for the Slow inward current component
        J_si = (
            -(1 + ufl.tanh(k * (-u_csi + (-V_0 + V) / (V_fi - V_0)))) * w / (2 * tau_si)
        )

        # Expressions for the Stimulus protocol component
        J_stim = 0

        # Expressions for the Membrane component
        current[0] = (V_0 - V_fi) * (J_stim + J_fi + J_si + J_so)

        # Return results
        return current[0]

    def I(self, v, s, time=None):
        """
        Transmembrane current

           I = -dV/dt

        """
        return -self._I(v, s, time)

    def F(self, v, s, time=None):
        """
        Right hand side for ODE system
        """
        time = time if time else backend.Constant(0.0)

        # Assign states
        V = v
        assert len(s) == 2
        v, w = s

        # Assign parameters
        u_c = self._parameters["u_c"]
        u_v = self._parameters["u_v"]
        tau_v1_minus = self._parameters["tau_v1_minus"]
        tau_v2_minus = self._parameters["tau_v2_minus"]
        tau_v_plus = self._parameters["tau_v_plus"]
        tau_w_minus = self._parameters["tau_w_minus"]
        tau_w_plus = self._parameters["tau_w_plus"]
        V_0 = self._parameters["V_0"]
        V_fi = self._parameters["V_fi"]

        # Init return args
        F_expressions = [ufl.zero()] * 2

        # Expressions for the p component
        p = ufl.conditional(ufl.lt((-V_0 + V) / (V_fi - V_0), u_c), 0, 1)

        # Expressions for the q component
        q = ufl.conditional(ufl.lt((-V_0 + V) / (V_fi - V_0), u_v), 0, 1)

        # Expressions for the v gate component
        tau_v_minus = tau_v1_minus * q + tau_v2_minus * (1 - q)
        F_expressions[0] = (1 - p) * (1 - v) / tau_v_minus - p * v / tau_v_plus

        # Expressions for the w gate component
        F_expressions[1] = (1 - p) * (1 - w) / tau_w_minus - p * w / tau_w_plus

        # Return results
        return dolfin.as_vector(F_expressions)

    def num_states(self):
        return 2

    def __str__(self):
        return "Fenton_karma_1998_BR_altered cardiac cell model"
